import argparse
import os

import onionbalance

TOR_CONTROL_SOCKET = os.environ.get('ONIONBALANCE_TOR_CONTROL_SOCKET',
                                    '/var/run/tor/control')


def get_common_argparser():
    """
    Parses and returns command line arguments.
    """

    parser = argparse.ArgumentParser(
        description="onionbalance distributes the requests for a Tor hidden "
        "services across multiple Tor instances.")

    parser.add_argument("--hs-version", type=str, default='v3',
                        choices=('v3', 'v2'),
                        help="Onion service version (default: v3)")

    parser.add_argument("-i", "--ip", type=str, default='127.0.0.1',
                        help="Tor controller IP address")

    parser.add_argument("-p", "--port", type=int, default=9051,
                        help="Tor controller port")

    parser.add_argument("-s", "--socket", type=str, default=TOR_CONTROL_SOCKET,
                        help="Tor unix domain control socket location")

    parser.add_argument("-c", "--config", type=str,
                        default=os.environ.get('ONIONBALANCE_CONFIG',
                                               "config.yaml"),
                        help="Config file location")

    parser.add_argument("-v", "--verbosity", type=str, default=None,
                        help="Minimum verbosity level for logging.  Available "
                             "in ascending order: debug, info, warning, "
                             "error, critical).  The default is info.")

    parser.add_argument('--version', action='version',
                        version='onionbalance %s' % onionbalance.__version__)

    parser.add_argument("--is-testnet", action='store_true',
                        help="Is this onionbalance on a test net? (Default: no)")

    return parser



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Global entry point to decide whether we are doing v2 or v3 so that we use the
right codebase
"""

from onionbalance.common import argparser
from onionbalance.common import log
from setproctitle import setproctitle  # pylint: disable=no-name-in-module

import onionbalance.hs_v2.manager
import onionbalance.hs_v3.manager

from onionbalance import __version__

logger = log.get_logger()


def main():
    setproctitle('onionbalance')
    parser = argparser.get_common_argparser()
    args = parser.parse_args()

    logger.warning("Initializing onionbalance (version: %s)...", __version__)

    if args.hs_version == 'v2':
        onionbalance.hs_v2.manager.main(args)
    else:
        onionbalance.hs_v3.manager.main(args)



import sys
import time
import os
import yaml
import stem

from stem.control import Controller

from onionbalance.common import log

logger = log.get_logger()


def read_config_data_from_file(config_path):
    if os.path.exists(config_path):
        with open(config_path, 'r') as handle:
            config_data = yaml.safe_load(handle.read())
            logger.info("Loaded the config file '%s'.", config_path)
    else:
        logger.error("The specified config file '%s' does not exist. The "
                     "onionbalance-config tool can generate the required "
                     "keys and config files.", config_path)
        sys.exit(1)

    return config_data


def connect_to_control_port(tor_socket=None, tor_address=None, tor_port=0, control_password=None):
    controller = None

    # Try first with a connection to the Tor unix domain control socket
    if tor_socket:
        try:
            controller = Controller.from_socket_file(path=tor_socket)
            logger.debug("Successfully connected to the Tor control socket "
                         "%s.", tor_socket)
        except stem.SocketError:
            logger.debug("Unable to connect to the Tor control socket %s.",
                         tor_socket)

    # If we didn't manage to connect to control socket, try IP:PORT
    if not controller:
        try:
            controller = Controller.from_port(address=tor_address,
                                              port=tor_port)
            logger.debug("Successfully connected to the Tor control port.")
        except stem.SocketError as exc:
            logger.error("Unable to connect to Tor control port: %s", exc)
            sys.exit(1)

    try:
        controller.authenticate(password=control_password)
    except stem.connection.AuthenticationFailure as exc:
        logger.error("Unable to authenticate on the Tor control connection: "
                     "%s", exc)
        sys.exit(1)
    else:
        logger.debug("Successfully authenticated on the Tor control "
                     "connection.")

    return controller


def reauthenticate(controller, logger, control_password=None):
    """
    Tries to authenticate to the controller
    """
    time.sleep(10)
    try:
        controller.authenticate(password=control_password)
    except stem.connection.AuthenticationFailure:
        logger.error("Failed to re-authenticate controller.")



# -*- coding: utf-8 -*-

"""
Implements the generation and loading of configuration files.
"""
import os
import sys
import errno

from onionbalance.hs_v2 import config
from onionbalance.common import log

import onionbalance.hs_v2.util
import onionbalance.common.util

import onionbalance.hs_v2.service
import onionbalance.hs_v2.instance

logger = log.get_logger()


def parse_config_file(config_file):
    """
    Parse config file containing service information
    """
    config_path = os.path.abspath(config_file)
    config_data = onionbalance.common.util.read_config_data_from_file(config_path)

    # Rewrite relative paths in the config to be relative to the config
    # file directory
    config_directory = os.path.dirname(config_path)
    for service in config_data.get('services'):
        if not os.path.isabs(service.get('key')):
            service['key'] = os.path.join(config_directory, service['key'])

    return config_data


def initialize_services(controller, services_config):
    """
    Load keys for services listed in the config
    """

    # Load the keys and config for each onion service
    for service in services_config:
        try:
            service_key = onionbalance.hs_v2.util.key_decrypt_prompt(service.get("key"))
        except ValueError as e:
            logger.error("Got exception '%s'. Perhaps you are trying to load a v3 onion service "
                         "but with --hs-version=v2 enabled?", e)
            sys.exit(1)
        except (IOError, OSError) as e:
            if e.errno == errno.ENOENT:
                logger.error("Private key file %s could not be found. "
                             "Relative paths in the config file are loaded "
                             "relative to the config file directory.",
                             service.get("key"))
                sys.exit(1)
            elif e.errno == errno.EACCES:
                logger.error("Permission denied to private key %s.",
                             service.get("key"))
                sys.exit(1)
            else:
                raise
        # Key file was read but a valid private key was not found.
        if not service_key:
            logger.error("Private key %s could not be loaded. It is a not "
                         "valid 1024 bit PEM encoded RSA private key",
                         service.get("key"))
            sys.exit(1)
        else:
            # Successfully imported the private key
            onion_address = onionbalance.hs_v2.util.calc_onion_address(service_key)
            logger.debug("Loaded private key for service %s.onion.",
                         onion_address)

        # Load all instances for the current onion service
        instance_config = service.get("instances", [])
        if not instance_config:
            logger.error("Could not load any instances for service "
                         "%s.onion.", onion_address)
            sys.exit(1)
        else:
            instances = []
            for instance in instance_config:
                instances.append(onionbalance.hs_v2.instance.InstanceV2(
                    controller=controller,
                    onion_address=instance.get("address"),
                    authentication_cookie=instance.get("auth")
                ))

            logger.info("Loaded %d instances for service %s.onion.",
                        len(instances), onion_address)

        # Store service configuration in config.services global
        config.services.append(onionbalance.hs_v2.service.Service(
            controller=controller,
            service_key=service_key,
            instances=instances
        ))

        # Store a global reference to current controller connection
        config.controller = controller



import time

import stem.control

from onionbalance.common import log
import onionbalance.common.util

logger = log.get_logger()


def helper_fetch_all_instance_descriptors(controller, instances, control_password=None):
    """
    Try fetch fresh descriptors for all HS instances
    """
    logger.info("Initiating fetch of descriptors for all service instances.")

    # pylint: disable=no-member

    while True:
        try:
            # Clear Tor descriptor cache before making fetches by sending
            # the NEWNYM singal
            controller.signal(stem.control.Signal.NEWNYM)
            time.sleep(5)  # Sleep to allow Tor time to build new circuits
            pass
        except stem.SocketClosed:
            logger.error("Failed to send NEWNYM signal, socket is closed.")
            onionbalance.common.util.reauthenticate(controller, logger, control_password)
        else:
            break

    unique_instances = set(instances)

    # Only try to retrieve the descriptor once for each unique instance
    # address. An instance may be configured under multiple master
    # addressed. We do not want to request the same instance descriptor
    # multiple times.
    # OnionBalance will update all of the matching instances when a
    # descriptor is received.
    for instance in unique_instances:
        while True:
            try:
                instance.fetch_descriptor()
            except stem.SocketClosed:
                logger.error("Failed to fetch descriptor, socket is closed")
                onionbalance.common.util.reauthenticate(controller, logger, control_password)
            else:
                break


class Instance(object):
    """
    Instance represents a back-end load balancing hidden service.
    """

    def __init__(self, controller, onion_address, authentication_cookie=None):
        """
        Initialise an Instance object.
        """
        self.controller = controller

        # Onion address for the service instance.
        if onion_address:
            onion_address = onion_address.replace('.onion', '')
        self.onion_address = onion_address

        # Flag this instance with its introduction points change. A new
        # master descriptor will then be published as the introduction
        # points have changed.
        self.intro_set_changed_since_published = False

    def fetch_descriptor(self):
        """
        Try fetch a fresh descriptor for this service instance from the HSDirs
        """
        logger.debug("Trying to fetch a descriptor for instance %s.onion.",
                     self.onion_address)
        try:
            self.controller.get_hidden_service_descriptor(self.onion_address,
                                                          await_result=False)
        except stem.SocketClosed:
            # Tor maybe restarting.
            raise
        except stem.DescriptorUnavailable:
            # Could not find the descriptor on the HSDir
            self.received = None
            logger.warning("No descriptor received for instance %s.onion, "
                           "the instance may be offline.", self.onion_address)

    def __eq__(self, other):
        """
        Instance objects are equal if they have the same onion address.
        """
        if isinstance(other, Instance):
            return self.onion_address == other.onion_address
        else:
            return False

    def __hash__(self):
        """
        Define __hash__ method allowing for set comparison between instances.
        """
        return hash(self.onion_address)



# -*- coding: utf-8 -*-
"""
Load balance a hidden service across multiple (remote) Tor instances by
create a hidden service descriptor containing introduction points from
each instance.
"""
import sys
import logging

# import Crypto.PublicKey
import stem
from stem.control import EventType

import onionbalance.common.util

from onionbalance.common import scheduler
from onionbalance.common import log
from onionbalance.common import signalhandler

from onionbalance.hs_v2 import config
from onionbalance.hs_v2 import eventhandler
from onionbalance.hs_v2 import settings
from onionbalance.hs_v2 import status

from onionbalance.hs_v2.service import publish_all_descriptors
from onionbalance.hs_v2.instance import fetch_instance_descriptors

logger = log.get_logger()


def main(args):
    """
    Entry point when invoked over the command line.
    """
    config_file_options = settings.parse_config_file(args.config)

    # Update global configuration with options specified in the config file
    for setting in dir(config):
        if setting.isupper() and config_file_options.get(setting):
            setattr(config, setting, config_file_options.get(setting))

    # Override the log level if specified on the command line.
    if args.verbosity:
        config.LOG_LEVEL = args.verbosity.upper()

    # Write log file if configured in environment variable or config file
    if config.LOG_LOCATION:
        log.setup_file_logger(config.LOG_LOCATION)

    logger.setLevel(logging.__dict__[config.LOG_LEVEL.upper()])

    controller = onionbalance.common.util.connect_to_control_port(args.socket, args.ip, args.port,
                                                                  config.TOR_CONTROL_PASSWORD)

    status_socket = status.StatusSocket(config.STATUS_SOCKET_LOCATION)
    signalhandler.SignalHandler(controller, status_socket)

    # Disable no-member due to bug with "Instance of 'Enum' has no * member"
    # pylint: disable=no-member

    # Check that the Tor client supports the HSPOST control port command
    if not controller.get_version() >= stem.version.Requirement.HSPOST:
        logger.error("A Tor version >= %s is required. You may need to "
                     "compile Tor from source or install a package from "
                     "the experimental Tor repository.",
                     stem.version.Requirement.HSPOST)
        sys.exit(1)

    # Load the keys and config for each onion service
    settings.initialize_services(controller,
                                 config_file_options.get('services'))

    # Finished parsing all the config file.

    handler = eventhandler.EventHandler()
    controller.add_event_listener(handler.new_status,
                                  EventType.STATUS_CLIENT)
    controller.add_event_listener(handler.new_desc,
                                  EventType.HS_DESC)
    controller.add_event_listener(handler.new_desc_content,
                                  EventType.HS_DESC_CONTENT)

    # Schedule descriptor fetch and upload events
    scheduler.add_job(config.REFRESH_INTERVAL, fetch_instance_descriptors,
                      controller)
    scheduler.add_job(config.PUBLISH_CHECK_INTERVAL, publish_all_descriptors)

    # Run initial fetch of HS instance descriptors
    scheduler.run_all(delay_seconds=config.INITIAL_DELAY)

    # Begin main loop to poll for HS descriptors
    scheduler.run_forever()

    return 0



# -*- coding: utf-8 -*-
"""
Provide status over Unix socket
Default path: /var/run/onionbalance/control
"""
import os
import errno
import threading
import socket
from socketserver import BaseRequestHandler, ThreadingMixIn, UnixStreamServer

from onionbalance.common import log
from onionbalance.hs_v2 import config

logger = log.get_logger()


class StatusSocketHandler(BaseRequestHandler):
    """
    Handler for new domain socket connections
    """

    def handle(self):
        """
        Prepare and output the status summary when a connection is received
        """
        time_format = "%Y-%m-%d %H:%M:%S"
        response = []
        for service in config.services:
            if service.uploaded:
                service_timestamp = service.uploaded.strftime(time_format)
            else:
                service_timestamp = "[not uploaded]"
            response.append("{}.onion {}".format(service.onion_address,
                                                 service_timestamp))

            for instance in service.instances:
                if not instance.timestamp:
                    response.append("  {}.onion [offline]".format(
                        instance.onion_address))
                else:
                    response.append("  {}.onion {} {} IPs".format(
                        instance.onion_address,
                        instance.timestamp.strftime(time_format),
                        len(instance.introduction_points)))
        response.append("")
        self.request.sendall('\n'.join(response).encode('utf-8'))


class ThreadingSocketServer(ThreadingMixIn, UnixStreamServer):
    """
    Unix socket server with threading
    """
    pass


class StatusSocket(object):
    """
    Create a Unix domain socket which emits a summary of the OnionBalance
    status when a client connects.
    """

    def __init__(self, status_socket_location):
        """
        Create the Unix domain socket status server and start in a thread

        Example::
            socat - unix-connect:/var/run/onionbalance/control

            uweyln7jhkyaokka.onion 2016-05-01 11:08:56
              r523s7jx65ckitf4.onion [offline]
              v2q7ujuleky7odph.onion 2016-05-01 11:00:00 3 IPs
        """
        self.unix_socket_filename = status_socket_location
        self.cleanup_socket_file()

        logger.debug("Creating status socket at %s", self.unix_socket_filename)
        try:
            self.server = ThreadingSocketServer(self.unix_socket_filename,
                                                StatusSocketHandler)

            # Start running the socket server in a another thread
            server_thread = threading.Thread(target=self.server.serve_forever)
            server_thread.daemon = True  # Exit daemon when main thread stops
            server_thread.start()

        except (OSError, socket.error):
            logger.error("Could not start status socket at %s. Does the path "
                         "exist? Do you have permission?",
                         status_socket_location)

    def cleanup_socket_file(self):
        """
        Try to remove the socket file if it exists already
        """
        try:
            os.unlink(self.unix_socket_filename)
        except OSError as e:
            # Reraise if its not a FileNotFound exception
            if e.errno != errno.ENOENT:
                raise

    def close(self):
        """
        Close the unix domain socket and remove its file
        """
        try:
            self.server.shutdown()
            self.server.server_close()
            self.cleanup_socket_file()
        except AttributeError:
            pass
        except OSError:
            logger.exception("Error when removing the status socket")



# -*- coding: utf-8 -*-

__author__ = "George Kadianakis, Donncha O'Cearbhaill"
__contact__ = "asn@torproject.org"
__url__ = "https://github.com/asn-d6/onionbalance"
__license__ = "GPL"

from onionbalance._version import get_versions
__version__ = get_versions()['version']
del get_versions



import datetime

import onionbalance.common.instance
from onionbalance.common import log

from onionbalance.hs_v3 import descriptor as ob_descriptor

logger = log.get_logger()


class InstanceV3(onionbalance.common.instance.Instance):
    """
    This is a V3 onionbalance instance
    """

    def __init__(self, onion_address):
        # Get the controller
        from onionbalance.hs_v3.onionbalance import my_onionbalance
        controller = my_onionbalance.controller.controller

        # Initialize the common Instance class.
        super().__init__(controller, onion_address)

        # Onion address does not contain the '.onion'.
        logger.warning("Loaded instance %s", onion_address)

        self.descriptor = None

        # When was the intro set of this instance last modified?
        self.intro_set_modified_timestamp = None

    def has_onion_address(self, onion_address):
        """
        Return True if this instance has this onion address
        """
        # Strip the ".onion" part of the address if it exists since some
        # subsystems don't use it (e.g. Tor sometimes omits it from control
        # port responses)
        my_onion_address = self.onion_address.replace(".onion", "")
        their_onion_address = onion_address.replace(".onion", "")

        return my_onion_address == their_onion_address

    def register_descriptor(self, descriptor_text, onion_address):
        """
        We received a descriptor (with 'descriptor_text') for 'onion_address'.
        Register it to this instance.
        """
        logger.info("Found instance %s for this new descriptor!", self.onion_address)

        assert(onion_address == self.onion_address)

        # Parse descriptor. If it parsed correctly, we know that this
        # descriptor is truly for this instance (since the onion address
        # matches)
        try:
            new_descriptor = ob_descriptor.ReceivedDescriptor(descriptor_text, onion_address)
        except ob_descriptor.BadDescriptor:
            logger.warning("Received bad descriptor for %s. Ignoring.", self.onion_address)
            return

        # Before replacing the current descriptor with this one, check if the
        # introduction point set changed:

        # If this is the first descriptor for this instance, the intro point set changed
        if not self.descriptor:
            logger.info("This is the first time we see a descriptor for instance %s!", self.onion_address)
            self.intro_set_modified_timestamp = datetime.datetime.utcnow()
            self.descriptor = new_descriptor
            return

        assert(self.descriptor)
        assert(new_descriptor.intro_set)

        # We already have a descriptor but this is a new one. Check the intro points!
        if new_descriptor.intro_set != self.descriptor.intro_set:
            logger.info("We got a new descriptor for instance %s and the intro set changed!", self.onion_address)
            self.intro_set_modified_timestamp = datetime.datetime.utcnow()
        else:
            logger.info("We got a new descriptor for instance %s but the intro set did not change.", self.onion_address)

        self.descriptor = new_descriptor

    def get_intros_for_publish(self):
        """
        Get a list of stem.descriptor.IntroductionPointV3 objects for this descriptor

        Raise :InstanceHasNoDescriptor: if there is no descriptor for this instance
        Raise :InstanceIsOffline: if the instance is offline.
        """
        if not self.descriptor:
            raise InstanceHasNoDescriptor

        if self.descriptor.is_old():
            raise InstanceIsOffline

        return self.descriptor.get_intro_points()


class InstanceHasNoDescriptor(Exception):
    pass


class InstanceIsOffline(Exception):
    pass



# -*- coding: utf-8 -*-
import logging
import logging.handlers

handler = logging.StreamHandler()
handler.setFormatter(logging.Formatter(fmt="%(asctime)s [%(levelname)s]: "
                                           "%(message)s"))

logger = logging.getLogger("onionbalance")
logger.addHandler(handler)
logger.setLevel(logging.DEBUG)


def get_logger():
    """
    Returns a logger.
    """
    return logger


def setup_file_logger(log_file):
    """
    Add log file handler to the existing logger
    """
    handler = logging.handlers.RotatingFileHandler(
        log_file, maxBytes=10485760, backupCount=3)
    handler.setFormatter(logging.Formatter(fmt="%(asctime)s [%(levelname)s]: "
                                           "%(message)s"))
    logging.getLogger('onionbalance').addHandler(handler)


def get_config_generator_logger():
    """
    Simplified logger for interactive config generator CLI
    """
    handler = logging.StreamHandler()
    handler.setFormatter(logging.Formatter(fmt="[%(levelname)s]: "
                                               "%(message)s"))

    logger = logging.getLogger("onionbalance-config")
    logger.addHandler(handler)
    logger.setLevel(logging.INFO)
    return logger



import argparse
import logging
import shutil
import os
import sys
import getpass
import yaml
import pkg_resources

import Crypto.PublicKey.RSA
from cryptography.hazmat.primitives.asymmetric.ed25519 import Ed25519PrivateKey
from cryptography.hazmat.primitives import serialization

from stem.descriptor.hidden_service import HiddenServiceDescriptorV3

import onionbalance
from onionbalance.common import log
from onionbalance.hs_v2 import util
from onionbalance.hs_v3 import tor_ed25519

# Simplify the logging output for the command line tool
logger = log.get_config_generator_logger()


class ConfigGenerator(object):
    def __init__(self, args, interactive):
        self.args = args
        self.interactive = interactive

        self.hs_version = None
        self.output_path = None
        self.master_key = None
        self.master_onion_address = None
        self.num_instances = None
        self.tag = None
        self.torrc_port_line = None
        self.instances = None
        self.master_dir = None
        self.config_file_path = None
        self.master_key_path = None

        # If this is set, it means that we read our private key from a tor
        # instance and hence we should just copy the private key over instead
        # of recreating the file (so that the Tor private key semantics are
        # maintained (see self.is_priv_key_in_tor_format in service.py)
        self.v3_loaded_key_from_file = False

        # Gather information required to create config file!
        self.gather_information()

        # Create config file!
        self.generate_config()

    def gather_information(self):
        self.hs_version = self.get_hs_version()
        assert(self.hs_version in ['v2', 'v3'])

        # Check if output directory exists, if not try create it
        self.output_path = self.get_output_path()

        if self.hs_version == "v2":
            self.master_dir = os.path.join(self.output_path, 'master')
        else:
            self.master_dir = self.output_path

        # Load the master key
        self.master_key, self.master_onion_address = self.load_master_key()

        # Finished loading/generating master key, now try generate keys for
        # each service instance
        self.num_instances, self.tag = self.get_num_instances()

        # Create HiddenServicePort line for instance torrc file
        if self.hs_version == 'v2':
            self.torrc_port_line = self.get_torrc_port_line()

        self.instances = self.create_instances()

    def generate_config(self):
        self.write_master_key_to_disk()

        assert(self.instances)

        # Create the onionbalance config file
        self.create_yaml_config_file()

        if self.hs_version == 'v2':
            # Generate config files for each service instance
            self.write_v2_instance_files()

            logger.info("Done! Successfully generated an OnionBalance config and %d "
                        "instance keys for service %s.onion.",
                        self.num_instances, self.master_onion_address)

        if self.hs_version == "v3":
            logger.info("Done! Successfully generated an OnionBalance config for service %s.onion.",
                        self.master_onion_address)
            logger.info("Now please edit '%s' with a text editor to add/remove/edit your backend instances.",
                        self.config_file_path)

    def get_output_path(self):
        """
        Get path to output directory and create if needed
        """
        output_path = None
        if self.interactive:
            output_path = input("Enter path to store generated config "
                                "[{}]: ".format(os.path.abspath(self.args.output)))
        output_path = output_path or self.args.output
        try:
            util.try_make_dir(output_path)
        except OSError:
            logger.exception("Problem encountered when trying to create the "
                             "output directory %s.", os.path.abspath(output_path))
            sys.exit(1)
        else:
            logger.debug("Created the output directory '%s'.",
                         os.path.abspath(output_path))

        # Do some directory validation
        if self.hs_version == 'v2' and not util.is_directory_empty(output_path):
            # The output directory should be empty to avoid having conflict keys
            # or config files.
            logger.error("The specified output directory '%s' is not empty. Please "
                         "delete any files and folders or specify another output "
                         "directory.", output_path)
            sys.exit(1)
        elif self.hs_version == 'v3':
            config_path = os.path.join(output_path, 'config.yaml')
            if os.path.isfile(config_path):
                logger.error("The specified output directory '%s' already contains a 'config.yaml' "
                             "file. Please clean the directory before starting config_generator.",
                             output_path)
                sys.exit(1)

        return output_path

    def get_hs_version(self):
        # Get the HS version
        hs_version = None
        if self.interactive:
            hs_version = input('Enter HS version ("v2" or "v3") (Leave empty for "v3"): ')
        hs_version = hs_version or self.args.hs_version

        if hs_version not in ["v2", "v3"]:
            logger.error('Only accepting "v2" and "v3" as HS versions')
            sys.exit(1)

        logger.info("Rolling with HS %s!", hs_version)

        return hs_version

    def load_master_key(self):
        """
        Return the key and onion address of the frontend service.
        """
        self.master_key_path = self.get_master_key_path()

        # master_key_path is now either None (if no key path is specified) or
        # set to the actual path
        if self.hs_version == 'v2':
            return self.load_v2_master_key(self.master_key_path)
        else:
            return self.load_v3_master_key(self.master_key_path)

    def get_master_key_path(self):
        # Load master key if specified
        master_key_path = None
        helper = " (i.e. path to 'hs_ed25519_secret_key')" if self.hs_version == 'v3' else ""
        if self.interactive:
            # Read key path from user
            master_key_path = input("Enter path to master service private key%s "
                                    "(Leave empty to generate a key): " % (helper))
        master_key_path = self.args.key or master_key_path

        # If a key path was specified make sure it exists
        if master_key_path:
            if not os.path.isfile(master_key_path):
                logger.error("The specified master service private key '%s' "
                             "could not be found. Please confirm the path and "
                             "file permissions are correct.", master_key_path)
                sys.exit(1)

        return master_key_path

    def _load_v3_master_key_from_file(self, master_key_path):
        """
        Load a private key straight from a Tor instance (no OBv3 keys supported)
        and return the private key and onion address.
        """
        try:
            with open(master_key_path, 'rb') as handle:
                pem_key_bytes = handle.read()
        except EnvironmentError as e:
            logger.error("Unable to read service private key file ('%s')", e)
            sys.exit(1)

        try:
            master_private_key = tor_ed25519.load_tor_key_from_disk(pem_key_bytes)
        except ValueError:
            logger.error("Please provide path to a valid Tor master key")
            sys.exit(1)
        identity_pub_key = master_private_key.public_key()
        identity_pub_key_bytes = identity_pub_key.public_bytes(encoding=serialization.Encoding.Raw,
                                                               format=serialization.PublicFormat.Raw)
        master_onion_address = HiddenServiceDescriptorV3.address_from_identity_key(identity_pub_key_bytes)

        # remove the trailing .onion
        master_onion_address = master_onion_address.replace(".onion", "")

        self.v3_loaded_key_from_file = True

        return master_private_key, master_onion_address

    def load_v3_master_key(self, master_key_path):
        if master_key_path: # load key from file
            return self._load_v3_master_key_from_file(master_key_path)
        else: # generate new v3 key
            master_private_key = Ed25519PrivateKey.generate()
            master_public_key = master_private_key.public_key()
            master_pub_key_bytes = master_public_key.public_bytes(encoding=serialization.Encoding.Raw,
                                                                  format=serialization.PublicFormat.Raw)
            master_onion_address = HiddenServiceDescriptorV3.address_from_identity_key(master_pub_key_bytes)
            # cut out the onion since that's what the rest of the code expects
            master_onion_address = master_onion_address.replace(".onion", "")

            return master_private_key, master_onion_address

    def load_v2_master_key(self, master_key_path):
        if master_key_path:
            # Try load the specified private key file
            master_key = util.key_decrypt_prompt(master_key_path)
            if not master_key:
                logger.error("The specified master private key %s could not "
                             "be loaded.", os.path.abspath(master_key))
                sys.exit(1)

            master_onion_address = util.calc_onion_address(master_key)
            logger.info("Successfully loaded a master key for service "
                        "%s.onion.", master_onion_address)
        else:
            # No key specified, begin generating a new one.
            master_key = Crypto.PublicKey.RSA.generate(1024)
            master_onion_address = util.calc_onion_address(master_key)
            logger.debug("Created a new master key for service %s.onion.",
                         master_onion_address)

        return master_key, master_onion_address

    def get_num_instances(self):
        """
        Get the number of instances and a tag name for them.
        """
        num_instances = None
        if self.interactive:
            limits = " (min: 1, max: 8)" if self.hs_version == "v3" else ""
            num_instances = input("Number of instance services to create (default: %d)%s: " %
                                  (self.args.num_instances, limits))
            # Cast to int if a number was specified
            try:
                num_instances = int(num_instances)
            except ValueError:
                num_instances = None
        num_instances = num_instances or self.args.num_instances
        logger.debug("Creating %d service instances.", num_instances)

        tag = None
        if self.interactive:
            tag = input("Provide a tag name to group these instances "
                        "[{}]: ".format(self.args.tag))
        tag = tag or self.args.tag

        return num_instances, tag

    def get_torrc_port_line(self):
        """
        Get the HiddenServicePort line for the instance torrc file
        """
        service_virtual_port = None
        if self.interactive:
            service_virtual_port = input("Specify the service virtual port (for "
                                         "client connections) [{}]: ".format(
                                             self.args.service_virtual_port))
        service_virtual_port = service_virtual_port or self.args.service_virtual_port

        service_target = None
        if self.interactive:
            # In interactive mode, change default target to match the specified
            # virtual port
            default_service_target = u'127.0.0.1:{}'.format(service_virtual_port)
            service_target = input("Specify the service target IP and port (where "
                                   "your service is listening) [{}]: ".format(
                                       default_service_target))
            service_target = service_target or default_service_target
        service_target = service_target or self.args.service_target
        torrc_port_line = u'HiddenServicePort {} {}'.format(service_virtual_port,
                                                            service_target)
        return torrc_port_line

    def create_instances(self):
        if self.hs_version == 'v2':
            return self.create_v2_instances()
        else:
            return self.create_v3_instances()

    def create_v2_instances(self):
        instances = []

        for i in range(0, self.num_instances):
            instance_key = Crypto.PublicKey.RSA.generate(1024)
            instance_address = util.calc_onion_address(instance_key)
            logger.debug("Created a key for instance %s.onion.",
                         instance_address)
            instances.append((instance_address, instance_key))

        return instances

    def create_v3_instances(self):
        instances = []

        for i in range(0, self.num_instances):
            instances.append(("<Enter the instance onion address here>", None))

        return instances

    def get_master_key_passphrase(self):
        # Get optional passphrase for master key
        # [TODO: Implement for v3]
        master_passphrase = None
        if self.interactive:
            master_passphrase = getpass.getpass(
                "Provide an optional password to encrypt the master private "
                "key (Not encrypted if no password is specified): ")
        return master_passphrase or self.args.password

    def write_master_key_to_disk(self):
        # Finished reading input, starting to write config files.
        util.try_make_dir(self.master_dir)
        master_key_file = os.path.join(self.master_dir,
                                       '{}.key'.format(self.master_onion_address))
        with open(master_key_file, "wb") as key_file:
            os.chmod(master_key_file, 384)  # chmod 0600 in decimal

            if self.hs_version == 'v2':
                master_passphrase = self.get_master_key_passphrase()
                key_file.write(self.master_key.exportKey(passphrase=master_passphrase))
            elif self.hs_version == 'v3' and self.v3_loaded_key_from_file:
                # If we loaded a v3 key from a file, copy the file directly
                # (see loaded_key_from_file comments).
                shutil.copyfile(self.master_key_path, master_key_file)
                logger.info("Copied v3 master key from %s to %s.",
                            self.master_key_path, master_key_file)
            else:
                # If we generated our own v3 master key, write it to file. If
                # 'master_key' does not exist, it means that we are loading it
                # from a file, so we dont need to write it to disk.
                master_key_formatted = self.master_key.private_bytes(encoding=serialization.Encoding.PEM,
                                                                     format=serialization.PrivateFormat.PKCS8,
                                                                     encryption_algorithm=serialization.NoEncryption())
                key_file.write(master_key_formatted)

            logger.debug("Successfully wrote master key to file %s.",
                         os.path.abspath(master_key_file))

    def write_v2_instance_files(self):
        for i, (instance_address, instance_key) in enumerate(self.instances):
            # Create a numbered directory for instance
            instance_dir = os.path.join(self.output_path, '{}{}'.format(self.tag, i + 1))
            instance_key_dir = os.path.join(instance_dir, instance_address)
            util.try_make_dir(instance_key_dir)
            os.chmod(instance_key_dir, 1472)  # chmod 2700 in decimal

            instance_key_file = os.path.join(instance_key_dir, 'private_key')
            with open(instance_key_file, "wb") as key_file:
                os.chmod(instance_key_file, 384)  # chmod 0600 in decimal
                key_file.write(instance_key.exportKey())
                logger.debug("Successfully wrote key for instance %s.onion to "
                             "file.", instance_address)

            # Write torrc file for each instance
            instance_torrc = os.path.join(instance_dir, 'instance_torrc')
            instance_torrc_template = pkg_resources.resource_string(
                __name__, 'data/torrc-instance-v2')
            with open(instance_torrc, "w") as torrc_file:
                torrc_file.write(instance_torrc_template.decode('utf-8'))
                # The ./ relative path prevents Tor from raising relative
                # path warnings. The relative path may need to be edited manual
                # to work on Windows systems.
                torrc_file.write(u"HiddenServiceDir {}\n".format(
                    instance_address))
                torrc_file.write(u"{}\n".format(self.torrc_port_line))

    def create_yaml_config_file(self):
        # Create YAML OnionBalance settings file for these instances
        service_data = {'key': '{}.key'.format(self.master_onion_address)}
        service_data['instances'] = [{'address': address,
                                      'name': '{}{}'.format(self.tag, i + 1)} for
                                     i, (address, _) in enumerate(self.instances)]
        settings_data = {'services': [service_data]}
        config_yaml = yaml.safe_dump(settings_data, default_flow_style=False)

        self.config_file_path = os.path.join(self.master_dir, 'config.yaml')
        with open(self.config_file_path, "w") as config_file:
            config_file.write(u"# OnionBalance Config File\n")
            config_file.write(config_yaml)
            logger.info("Wrote master service config file '%s'.",
                        os.path.abspath(self.config_file_path))

        if self.hs_version == 'v2':
            # Write frontend service torrc
            master_torrc_path = os.path.join(self.master_dir, 'torrc-server')
            master_torrc_template = pkg_resources.resource_string(__name__,
                                                                  'data/torrc-server')
            with open(master_torrc_path, "w") as master_torrc_file:
                master_torrc_file.write(master_torrc_template.decode('utf-8'))


def parse_cmd_args():
    """
    Parses and returns command line arguments for config generator
    """

    parser = argparse.ArgumentParser(
        description="onionbalance-config generates config files and keys for "
        "OnionBalance instances and management servers. Calling without any "
        "options will initiate an interactive mode.")

    parser.add_argument("--hs-version", type=str,
                        default="v3",
                        help="Onion service version (default: %(default)s).")

    parser.add_argument("--key", type=str, default=None,
                        help="RSA private key for the master onion service.")

    parser.add_argument("-p", "--password", type=str, default=None,
                        help="Optional password which can be used to encrypt"
                        "the master service private key.")

    parser.add_argument("-n", type=int, default=2, dest="num_instances",
                        help="Number of instances to generate (default: "
                        "%(default)s).")

    parser.add_argument("-t", "--tag", type=str, default='node',
                        help="Prefix name for the service instances "
                        "(default: %(default)s).")

    parser.add_argument("--output", type=str, default='config/',
                        help="Directory to store generate config files. "
                        "The directory will be created if it does not "
                        "already exist.")

    parser.add_argument("--no-interactive", action='store_true',
                        help="Try to run automatically without prompting for"
                        "user input.")

    parser.add_argument("-v", type=str, default="info", dest='verbosity',
                        help="Minimum verbosity level for logging. Available "
                        "in ascending order: debug, info, warning, error, "
                        "critical).  The default is info.")

    parser.add_argument("--service-virtual-port", type=str,
                        default="80",
                        help="Onion service port for external client "
                        "connections (default: %(default)s).")

    # TODO: Add validator to check if the target host:port line makes sense.
    parser.add_argument("--service-target", type=str,
                        default="127.0.0.1:80",
                        help="Target IP and port where your service is "
                        "listening (default: %(default)s).")

    # .. todo:: Add option to specify HS host and port for instance torrc

    parser.add_argument('--version', action='version',
                        version='onionbalance %s' % onionbalance.__version__)

    return parser


def main():
    """
    Entry point for interactive config file generation.
    """

    # Parse initial command line options
    args = parse_cmd_args().parse_args()

    logger.info("Beginning OnionBalance config generation.")

    # If CLI options have been provided, don't enter interactive mode
    # Crude check to see if any options beside --verbosity are set.
    verbose = True if '-v' in sys.argv else False

    if ((len(sys.argv) > 1 and not verbose) or len(sys.argv) > 3 or args.no_interactive):
        interactive = False
        logger.info("Entering non-interactive mode.")
    else:
        interactive = True
        logger.info("No command line arguments found, entering interactive "
                    "mode.")

    logger.setLevel(logging.__dict__[args.verbosity.upper()])

    # Start the config generator!
    try:
        ConfigGenerator(args, interactive)
    except KeyboardInterrupt:
        logger.warning("\nConfig generator got interrupted! There might be temporary configuration files left over... Bye!")

    sys.exit(0)



# -*- coding: utf-8 -*-
import hashlib
import struct
import datetime
import getpass
import base64
import binascii
import os

# import Crypto.Util
import Crypto.PublicKey


def add_pkcs1_padding(message):
    """Add PKCS#1 padding to **message**."""
    padding = b''
    typeinfo = b'\x00\x01'
    separator = b'\x00'
    padding = b'\xFF' * (125 - len(message))
    padded_message = typeinfo + padding + separator + message
    assert len(padded_message) == 128
    return padded_message


def get_asn1_sequence(rsa_key):
    seq = Crypto.Util.asn1.DerSequence()
    seq.append(rsa_key.n)
    seq.append(rsa_key.e)
    asn1_seq = seq.encode()
    return asn1_seq


def calc_key_digest(rsa_key):
    """Calculate the SHA1 digest of an RSA key"""
    return hashlib.sha1(get_asn1_sequence(rsa_key)).digest()


def calc_permanent_id(rsa_key):
    return calc_key_digest(rsa_key)[:10]


def calc_onion_address(rsa_key):
    return base64.b32encode(calc_permanent_id(rsa_key)).decode().lower()


def calc_descriptor_id(permanent_id, secret_id_part):
    return hashlib.sha1(permanent_id + secret_id_part).digest()


def get_time_period(time, permanent_id):
    """
    time-period = (current-time + permanent-id-byte * 86400 / 256) / 86400
    """
    permanent_id_byte = int(struct.unpack('B', permanent_id[0:1])[0])
    return int((int(time) + permanent_id_byte * 86400 / 256) / 86400)


def get_seconds_valid(time, permanent_id):
    """
    Calculate seconds until the descriptor ID changes
    """
    permanent_id_byte = int(struct.unpack('B', permanent_id[0:1])[0])
    return 86400 - int((int(time) + permanent_id_byte * 86400 / 256) % 86400)


def calc_secret_id_part(time_period, descriptor_cookie, replica):
    """
    secret-id-part = H(time-period | descriptor-cookie | replica)
    """
    secret_id_part = hashlib.sha1()
    secret_id_part.update(struct.pack('>I', time_period)[:4])
    if descriptor_cookie:
        secret_id_part.update(descriptor_cookie)
    secret_id_part.update(binascii.unhexlify('{0:02X}'.format(replica)))
    return secret_id_part.digest()


def calc_descriptor_id_b32(onion_address, time, replica, deviation=0,
                           descriptor_cookie=None):
    """
    High level function to calculate the descriptor ID for a hidden
    service.

    The onion address and returned descriptor ID are both base32 encoded.
    """
    permanent_id = base64.b32decode(onion_address, 1)
    time_period = get_time_period(time, permanent_id) + int(deviation)
    secret_id_part = calc_secret_id_part(time_period, descriptor_cookie,
                                         replica)
    descriptor_id = calc_descriptor_id(permanent_id, secret_id_part)

    return base64.b32encode(descriptor_id).decode('utf-8').lower()


def rounded_timestamp(timestamp=None):
    """
    Create timestamp rounded down to the nearest hour
    """
    if not timestamp:
        timestamp = datetime.datetime.utcnow()
    timestamp = timestamp.replace(minute=0, second=0, microsecond=0)
    return timestamp.strftime('%Y-%m-%d %H:%M:%S')


def base32_encode_str(byte_str):
    """
    Encode bytes as lowercase base32 string
    """
    return base64.b32encode(byte_str).lower().decode('utf-8')


def key_decrypt_prompt(key_file, retries=3):
    """
    Try open an PEM encrypted private key, prompting the user for a
    passphrase if required.
    """

    key_passphrase = None
    with open(key_file, 'rt') as handle:
        pem_key = handle.read()

        for retries in range(0, retries):
            if "Proc-Type: 4,ENCRYPTED" in pem_key:  # Key looks encrypted
                key_passphrase = getpass.getpass(
                    "Enter the password for the private key (%s): " % key_file)
            try:
                rsa_key = Crypto.PublicKey.RSA.importKey(
                    pem_key, passphrase=key_passphrase)
            except ValueError:
                # Key not decrypted correctly, prompt for passphrase again
                continue
            else:
                # .. todo:: Check the loaded key size in a more reasonable way.
                if rsa_key.has_private() and rsa_key.size() in (1023, 1024):
                    return rsa_key
                else:
                    raise ValueError("The specified key was not a 1024 bit "
                                     "private key.")

    # No private key was imported
    raise ValueError("Could not import RSA key.")


def try_make_dir(path):
    """
    Try to create a directory (including any parent directories)
    """
    try:
        os.makedirs(path)
    except OSError:
        if not os.path.isdir(path):
            raise


def is_directory_empty(path):
    """
    Check if a directory contains any files or directories.
    """
    if os.listdir(path):
        return False
    else:
        return True



from future.moves.itertools import zip_longest
import random
import itertools


class IntroductionPointSet(object):
    """
    A set of introduction points to included in a HS descriptor.

    Provided with a list of available introduction points for each backend
    instance for an onionbalance service. This object will store the set of
    available introduction points and allow IPs to be selected from the
    available set.

    This class tracks which introduction points have already been provided
    and tries to provide the most diverse set of IPs.
    """

    def __init__(self, intro_points):
        """
        'intro_points' is a list of lists that looks like this:
        [
          [<intro#1 of Instance#1, intro#2 of Instance#1...>],
          [<intro#1 of Instance#2, intro#2 of Instance#2...>],
          [<intro#1 of Instance#3, intro#2 of Instance#3...>],
          ...
        ]
        """
        # Shuffle the introduction point order before selecting IPs.
        # Randomizing now allows later calls to .choose() to be
        # deterministic.
        for instance_intro_points in intro_points:
            random.shuffle(instance_intro_points)
        random.shuffle(intro_points)

        self.intro_points = intro_points
        self._intro_point_generator = self._get_intro_point()

    def __len__(self):
        """Provide the total number of available introduction points"""
        return sum(len(ips) for ips in self.intro_points)

    def _get_intro_point(self):
        """
        [Private function]

        Generator function which yields an introduction point

        Iterates through all available introduction points and try
        to pick IPs breath first across all backend instances. The
        intro point set is wrapped in `itertools.cycle` and will provided
        an infinite series of introduction points.
        """

        # Combine intro points from across the backend instances and flatten
        intro_points = zip_longest(*self.intro_points)
        flat_intro_points = itertools.chain.from_iterable(intro_points)
        for intro_point in itertools.cycle(flat_intro_points):
            if intro_point:
                yield intro_point

    def choose(self, count=10, shuffle=True):
        """
        [Public API]

        Retrieve N introduction points from the set of IPs

        Where more than `count` IPs are available, introduction points are
        selected to try and achieve the greatest distribution of introduction
        points across all of the available backend instances.

        Return a list of IntroductionPoints.
        """

        # Limit `count` to the available number of IPs to avoid repeats.
        count = min(len(self), count)
        choosen_ips = list(itertools.islice(self._intro_point_generator, count))

        if shuffle:
            random.shuffle(choosen_ips)
        return choosen_ips



import os
import sys

import stem
from stem.descriptor.hidden_service import HiddenServiceDescriptorV3

import onionbalance.common.instance

from onionbalance.common import log

from onionbalance.common import util

from onionbalance.hs_v3 import stem_controller
from onionbalance.hs_v3 import service as ob_service
from onionbalance.hs_v3 import consensus as ob_consensus

logger = log.get_logger()


class Onionbalance(object):
    """
    Onionbalance singleton that represents this onionbalance runtime.

    Contains various objects that are useful to other onionbalance modules so
    this is imported from all over the codebase.
    """

    def __init__(self):
        # This is kept minimal so that it's quick (it's executed at program
        # launch because of the onionbalance singleton). The actual init work
        # happens in init_subsystems()

        # True if this onionbalance operates in a testnet (e.g. chutney)
        self.is_testnet = False

    def init_subsystems(self, args):
        """
        Initialize subsystems (this is resource intensive)
        """
        self.args = args
        self.config_path = os.path.abspath(self.args.config)
        self.config_data = self.load_config_file()
        self.is_testnet = args.is_testnet

        if self.is_testnet:
            logger.warning("Onionbalance configured on a testnet!")

        # Create stem controller and connect to the Tor network
        self.controller = stem_controller.StemController(args.ip, args.port)
        self.consensus = ob_consensus.Consensus()

        # Initialize our service
        self.services = self.initialize_services_from_config_data()

        # Catch interesting events (like receiving descriptors etc.)
        self.controller.add_event_listeners()

        logger.warning("Onionbalance initialized (stem version: %s) (tor version: %s)!",
                       stem.__version__, self.controller.controller.get_version())
        logger.warning("=" * 80)

    def initialize_services_from_config_data(self):
        services = []
        try:
            for service in self.config_data['services']:
                services.append(ob_service.OnionBalanceService(service, self.config_path))
        except ob_service.BadServiceInit:
            sys.exit(1)

        if len(services) > 1:
            # We don't know how to handle more than a single service right now
            raise NotImplementedError

        return services

    def load_config_file(self):
        config_data = util.read_config_data_from_file(self.config_path)
        logger.debug("Onionbalance config data: %s", config_data)

        # Do some basic validation
        if "services" not in config_data:
            logger.error("Config file is bad. 'services' is missing. Did you make it with onionbalance-config?")
            sys.exit(1)

        # More validation
        for service in config_data["services"]:
            if "key" not in service:
                logger.error("Config file is bad. 'key' is missing. Did you make it with onionbalance-config?")
                sys.exit(1)

            if "instances" not in service:
                logger.error("Config file is bad. 'instances' is missing. Did you make it with onionbalance-config?")
                sys.exit(1)

            if not service["instances"]:
                logger.error("Config file is bad. No backend instances are set. Onionbalance needs at least 1.")
                sys.exit(1)

            for instance in service["instances"]:
                if "address" not in instance:
                    logger.error("Config file is wrong. 'address' missing from instance.")
                    sys.exit(1)

                if not instance["address"]:
                    logger.error("Config file is bad. Address field is not set.")
                    sys.exit(1)

                # Validate that the onion address is legit
                try:
                    _ = HiddenServiceDescriptorV3.identity_key_from_address(instance["address"])
                except ValueError:
                    logger.error("Cannot load instance with address: '%s'.", instance["address"])
                    logger.error("If you are trying to run onionbalance for v2 onions, please use the --hs-version=v2 switch")
                    sys.exit(1)

        return config_data

    def fetch_instance_descriptors(self):
        logger.info("[*] fetch_instance_descriptors() called [*]")

        # TODO: Don't do this here. Instead do it on a specialized function
        self.controller.mark_tor_as_active()

        if not self.consensus.is_live():
            logger.warning("No live consensus. Waiting before fetching descriptors...")
            return

        all_instances = self._get_all_instances()

        onionbalance.common.instance.helper_fetch_all_instance_descriptors(self.controller.controller,
                                                                           all_instances)

    def handle_new_desc_content_event(self, desc_content_event):
        """
        Parse HS_DESC_CONTENT response events for descriptor content

        Update the HS instance object with the data from the new descriptor.
        """
        onion_address = desc_content_event.address
        logger.debug("Received descriptor for %s.onion from %s",
                     onion_address, desc_content_event.directory)

        #  Check that the HSDir returned a descriptor that is not empty
        descriptor_text = str(desc_content_event.descriptor).encode('utf-8')

        # HSDirs provide a HS_DESC_CONTENT response with either one or two
        # CRLF lines when they do not have a matching descriptor. Using
        # len() < 5 should ensure all empty HS_DESC_CONTENT events are matched.
        if len(descriptor_text) < 5:
            logger.debug("Empty descriptor received for %s.onion", onion_address)
            return None

        # OK this descriptor seems plausible: Find the instances that this
        # descriptor belongs to:
        for instance in self._get_all_instances():
            if instance.onion_address == onion_address:
                instance.register_descriptor(descriptor_text, onion_address)

    def publish_all_descriptors(self):
        """
        For each service attempt to publish all descriptors
        """
        logger.info("[*] publish_all_descriptors() called [*]")

        if not self.consensus.is_live():
            logger.info("No live consensus. Waiting before publishing descriptors...")
            return

        for service in self.services:
            service.publish_descriptors()

    def _get_all_instances(self):
        """
        Get all instances for all services
        """
        instances = []

        for service in self.services:
            instances.extend(service.instances)

        return instances

    def handle_new_status_event(self, status_event):
        """
        Parse Tor status events such as "STATUS_GENERAL"
        """
        # pylint: disable=no-member
        if status_event.action == "CONSENSUS_ARRIVED":
            logger.info("Received new consensus!")
            self.consensus.refresh()
            # Call all callbacks in case we just got a live consensus
            my_onionbalance.publish_all_descriptors()
            my_onionbalance.fetch_instance_descriptors()

    def _address_is_instance(self, onion_address):
        """
        Return True if 'onion_address' is one of our instances.
        """
        for service in self.services:
            for instance in service.instances:
                if instance.has_onion_address(onion_address):
                    return True
        return False

    def _address_is_frontend(self, onion_address):
        for service in self.services:
            if service.has_onion_address(onion_address):
                return True
        return False

    def handle_new_desc_event(self, desc_event):
        """
        Parse HS_DESC response events
        """
        action = desc_event.action

        if action == "RECEIVED":
            pass  # We already log in HS_DESC_CONTENT so no need to do it here too
        elif action == "UPLOADED":
            logger.debug("Successfully uploaded descriptor for %s to %s", desc_event.address, desc_event.directory)
        elif action == "FAILED":
            if self._address_is_instance(desc_event.address):
                logger.info("Descriptor fetch failed for instance %s from %s (%s)",
                            desc_event.address, desc_event.directory, desc_event.reason)
            elif self._address_is_frontend(desc_event.address):
                logger.warning("Descriptor upload failed for frontend %s to %s (%s)",
                               desc_event.address, desc_event.directory, desc_event.reason)
            else:
                logger.warning("Descriptor action failed for unknown service %s to %s (%s)",
                               desc_event.address, desc_event.directory, desc_event.reason)
        elif action == "REQUESTED":
            logger.debug("Requested descriptor for %s from %s...", desc_event.address, desc_event.directory)
        else:
            pass


my_onionbalance = Onionbalance()



# -*- coding: utf-8 -*-
"""
This module provides a set of functions for reading Tor consensus documents.
"""
from bisect import bisect_left
import base64
import binascii

import stem
import stem.descriptor

import onionbalance.common.log as log
import onionbalance.hs_v2.config as config

logger = log.get_logger()

HSDIR_LIST = []


def refresh_consensus():
    """
    Update consensus state when Tor receives a new network status

    Retrieve the current set of hidden service directories
    """
    global HSDIR_LIST

    if not config.controller:
        logger.warning("Controller connection not found in the configuration. "
                       "Cannot update the Tor state.")
        return None
    else:
        controller = config.controller

    # pylint: disable=no-member
    # Retrieve the current set of hidden service directories
    hsdirs = []
    try:
        for desc in controller.get_network_statuses():
            if stem.Flag.HSDIR in desc.flags:
                hsdirs.append(desc.fingerprint)
    except IOError as err:
        logger.error("Could not load consensus from Tor: %s" % err)
    else:
        HSDIR_LIST = hsdirs
        logger.debug("Updated the list of Tor hidden service directories.")


def get_hsdirs(descriptor_id):
    """
    Get the responsible HSDirs for a given descriptor ID.
    """

    # Try fetch a consensus if we haven't loaded one already
    if not HSDIR_LIST:
        refresh_consensus()

    if not HSDIR_LIST:
        raise ValueError('Could not determine the responsible HSDirs.')

    desc_id_bytes = base64.b32decode(descriptor_id, 1)
    descriptor_id_hex = (binascii.hexlify(desc_id_bytes).
                         decode('utf-8').upper())

    responsible_hsdirs = []

    # Find postion of descriptor ID in the HSDir list
    index = descriptor_position = bisect_left(HSDIR_LIST, descriptor_id_hex)

    # Pick HSDirs until we have enough
    while len(responsible_hsdirs) < config.HSDIR_SET:
        try:
            responsible_hsdirs.append(HSDIR_LIST[index])
            index += 1
        except IndexError:
            # Wrap around when we reach the end of the HSDir list
            index = 0

        # Do not choose a HSDir more than once
        if index == descriptor_position:
            break

    return responsible_hsdirs



import datetime
import base64
import hashlib

import stem
import stem.util
import stem.descriptor.remote
from stem.descriptor.networkstatus import NetworkStatusDocumentV3

from onionbalance.common import log
from onionbalance.hs_v3 import tor_node

logger = log.get_logger()


class Consensus(object):
    """
    This represents a consensus object.

    It's initialized once in startup and refreshed during the runtime using the
    refresh() method to get the latest consensus.
    """

    def __init__(self, do_refresh_consensus=True):
        # A list of tor_node:Node objects contained in the current consensus
        self.nodes = None
        # A stem NetworkStatusDocumentV3 object representing the current consensus
        self.consensus = None

        if not do_refresh_consensus:
            return

        # Set self.consensus to a NetworkStatusDocumentV3 object
        # and initialize the nodelist
        self.refresh()

    def refresh(self):
        """
        Attempt to refresh the consensus with the latest one available.
        """
        from onionbalance.hs_v3.onionbalance import my_onionbalance

        # Fetch the current md consensus from the control port
        md_consensus_str = my_onionbalance.controller.get_md_consensus().encode()
        try:
            self.consensus = NetworkStatusDocumentV3(md_consensus_str)
        except ValueError:
            logger.warning("No valid consensus received. Waiting for one...")
            return

        # Check if it's live
        if not self.is_live():
            logger.info("Loaded consensus is not live. Waiting for a live one.")
            return

        self.nodes = self._initialize_nodes()

    def get_routerstatuses(self):
        """Give access to the routerstatus entries in this consensus"""

        # We should never be asked for routerstatuses with a non-live consensus
        # so make sure this is the case.
        assert(self.is_live())

        return self.consensus.routers

    def is_live(self):
        """
        Return True if the consensus is live.
        """
        if not self.consensus:
            return False

        now = datetime.datetime.utcnow()
        return self.consensus.valid_after <= now and now <= self.consensus.valid_until

    def _initialize_nodes(self):
        """
        Initialize self.nodes with the list of current nodes.
        """
        from onionbalance.hs_v3.onionbalance import my_onionbalance

        nodes = []

        try:
            microdescriptors_list = list(my_onionbalance.controller.controller.get_microdescriptors())
        except stem.DescriptorUnavailable:
            logger.warning("Can't get microdescriptors from Tor. Delaying...")
            return

        # Turn the mds into a dictionary indexed by the digest as an
        # optimization while matching them with routerstatuses.
        microdescriptors_dict = {}
        for md in microdescriptors_list:
            microdescriptors_dict[md.digest()] = md

        # Go through the routerstatuses and match them up with
        # microdescriptors, and create a Node object for each match. If there
        # is no match we don't register it as a node.
        for relay_fpr, relay_routerstatus in self.get_routerstatuses().items():
            logger.debug("Checking routerstatus with md digest %s", relay_routerstatus.microdescriptor_digest)

            # Skip routerstatuses for which we cannot find a microdescriptor
            if relay_routerstatus.microdescriptor_digest not in microdescriptors_dict:
                logger.debug("Could not find microdesc for rs with fpr %s", relay_fpr)
                continue

            node_microdescriptor = microdescriptors_dict[relay_routerstatus.microdescriptor_digest]
            node = tor_node.Node(node_microdescriptor, relay_routerstatus)
            nodes.append(node)

        logger.debug("Initialized %d nodes (%d routerstatuses / %d microdescriptors)",
                     len(nodes), len(self.get_routerstatuses()), len(microdescriptors_list))

        return nodes

    def _get_disaster_srv(self, time_period_num):
        """
        Return disaster SRV for 'time_period_num'.
        """
        time_period_length = self.get_time_period_length()

        disaster_body = b"shared-random-disaster" + time_period_length.to_bytes(8, 'big') + time_period_num.to_bytes(8, 'big')
        return hashlib.sha3_256(disaster_body).digest()

    def get_current_srv(self, time_period_num):
        if self.consensus.shared_randomness_current_value:
            return base64.b64decode(self.consensus.shared_randomness_current_value)
        elif time_period_num:
            logger.info("SRV not found so falling back to disaster mode")
            return self._get_disaster_srv(time_period_num)
        else:
            return None

    def get_previous_srv(self, time_period_num):
        if self.consensus.shared_randomness_previous_value:
            return base64.b64decode(self.consensus.shared_randomness_previous_value)
        elif time_period_num:
            logger.info("SRV not found so falling back to disaster mode")
            return self._get_disaster_srv(time_period_num)
        else:
            return None

    def _get_srv_phase_duration(self):
        """
        Return the length of the phase of a shared random protocol run in minutes.
        """

        # Each SRV phase takes 12 rounds. But the duration of the round depends
        # on how big the voting rounds are which differs between live and
        # testing network:
        from onionbalance.hs_v3.onionbalance import my_onionbalance
        if my_onionbalance.is_testnet:
            return (12 * 20) // 60
        else:
            return 12 * 60

    def get_time_period_length(self):
        """
        Get the HSv3 time period length in minutes
        """
        from onionbalance.hs_v3.onionbalance import my_onionbalance
        if my_onionbalance.is_testnet:
            # This is a chutney network! Use hs_common.c:get_time_period_length()
            # logic to calculate time period length
            return (24 * 20) // 60
        else:
            # This is not a chutney network, so time period length is 1440 minutes (24 hours)
            return 24 * 60

    def get_blinding_param(self, identity_pubkey, time_period_number):
        """
        Calculate the HSv3 blinding parameter as specified in rend-spec-v3.txt section A.2:

         h = H(BLIND_STRING | A | s | B | N)
         BLIND_STRING = "Derive temporary signing key" | INT_1(0)
         N = "key-blind" | INT_8(period-number) | INT_8(period_length)
         B = "(1511[...]2202, 4631[...]5960)"

        Use the time period number in 'time_period_number'.
        """
        ED25519_BASEPOINT = b"(15112221349535400772501151409588531511" \
            b"454012693041857206046113283949847762202, " \
            b"463168356949264781694283940034751631413" \
            b"07993866256225615783033603165251855960)"
        BLIND_STRING = b"Derive temporary signing key" + bytes([0])

        period_length = self.get_time_period_length()
        N = b"key-blind" + time_period_number.to_bytes(8, 'big') + period_length.to_bytes(8, 'big')

        return hashlib.sha3_256(BLIND_STRING + identity_pubkey + ED25519_BASEPOINT + N).digest()

    def get_next_time_period_num(self, valid_after=None):
        return self.get_time_period_num(valid_after) + 1

    def get_time_period_num(self, valid_after=None):
        """
        Get time period number for this 'valid_after'.

        valid_after is a datetime (if not set, we get it ourselves)
        time_period_length set to default value of 1440 minutes == 1 day
        """
        if not valid_after:
            assert(self.is_live())
            valid_after = self.consensus.valid_after
            valid_after = stem.util.datetime_to_unix(valid_after)

        time_period_length = self.get_time_period_length()

        seconds_since_epoch = valid_after
        minutes_since_epoch = seconds_since_epoch // 60

        # Calculate offset as specified in rend-spec-v3.txt [TIME-PERIODS]
        time_period_rotation_offset = self._get_srv_phase_duration()

        assert(minutes_since_epoch > time_period_rotation_offset)
        minutes_since_epoch -= time_period_rotation_offset

        time_period_num = minutes_since_epoch // time_period_length
        return int(time_period_num)

    def get_start_time_of_current_srv_run(self):
        """
        Return the start time of the current SR protocol run using the times from
        the current consensus. For example, if the latest consensus valid-after is
        23/06/2017 23:00:00 and a full SR protocol run is 24 hours, this function
        returns 23/06/2017 00:00:00.

        TODO: unittest
        """
        from onionbalance.hs_v3.onionbalance import my_onionbalance

        assert(self.is_live())

        beginning_of_current_round = stem.util.datetime_to_unix(self.consensus.valid_after)

        # Voting interval is 20 secs in chutney and one hour in real network
        if my_onionbalance.is_testnet:
            voting_interval_secs = 20
        else:
            voting_interval_secs = 60 * 60

        # Get current SR protocol round (an SR protocol run has 24 rounds)
        curr_round_slot = (beginning_of_current_round // voting_interval_secs) % 24
        time_elapsed_since_start_of_run = curr_round_slot * voting_interval_secs

        logger.debug("Current SRV proto run: Start of current round: %s. "
                     "Time elapsed: %s (%s)", beginning_of_current_round,
                     time_elapsed_since_start_of_run, voting_interval_secs)

        return int(beginning_of_current_round - time_elapsed_since_start_of_run)

    def get_start_time_of_previous_srv_run(self):
        from onionbalance.hs_v3.onionbalance import my_onionbalance

        start_time_of_current_run = self.get_start_time_of_current_srv_run()
        if my_onionbalance.is_testnet:
            return start_time_of_current_run - 24 * 20
        else:
            return start_time_of_current_run - 24 * 3600

    def get_start_time_of_next_time_period(self, valid_after=None):
        """
        Return the start time of the upcoming time period
        """
        assert(self.is_live())

        # Get start time of next time period
        time_period_length = self.get_time_period_length()
        next_time_period_num = self.get_next_time_period_num(valid_after)
        start_of_next_tp_in_mins = next_time_period_num * time_period_length

        # Apply rotation offset as specified by prop224 section [TIME-PERIODS]
        time_period_rotation_offset = self._get_srv_phase_duration()

        return (start_of_next_tp_in_mins + time_period_rotation_offset) * 60


class NoLiveConsensus(Exception):
    pass



import traceback

import stem
from stem.control import EventType
from stem import Signal

import onionbalance.common.util
from onionbalance.common import log

logger = log.get_logger()


def handle_new_status_event_wrapper(status_event):
    """
    A wrapper for this control port event. We need this so that we print
    tracebacks on the listener thread (also see
    https://stem.torproject.org/tutorials/tortoise_and_the_hare.html#advanced-listeners)
    """
    from onionbalance.hs_v3.onionbalance import my_onionbalance
    try:
        my_onionbalance.handle_new_status_event(status_event)
    except BaseException:
        print(traceback.format_exc())


def handle_new_desc_event_wrapper(desc_event):
    """  A wrapper for this control port event (see above) """
    from onionbalance.hs_v3.onionbalance import my_onionbalance
    try:
        my_onionbalance.handle_new_desc_event(desc_event)
    except BaseException:
        print(traceback.format_exc())


def handle_new_desc_content_event_wrapper(desc_content_event):
    """  A wrapper for this control port event (see above) """
    from onionbalance.hs_v3.onionbalance import my_onionbalance
    try:
        my_onionbalance.handle_new_desc_content_event(desc_content_event)
    except BaseException:
        print(traceback.format_exc())


class StemController(object):
    """This class is our interface to the control port"""

    def __init__(self, address, port):
        self.controller = onionbalance.common.util.connect_to_control_port(tor_address=address,
                                                                           tor_port=port)
        assert(self.controller.is_authenticated())

    def mark_tor_as_active(self):
        """
        Send the ACTIVE signal to the control port so that Tor does not become dormant.
        """
        # pylint: disable=no-member
        try:
            self.controller.signal(Signal.ACTIVE)
        except stem.SocketClosed:
            logger.warning("Can't connect to the control port to send ACTIVE signal. Moving on...")

    def get_md_consensus(self):
        return self.controller.get_info("dir/status-vote/current/consensus-microdesc")

    def add_event_listeners(self):
        # pylint: disable=no-member

        self.controller.add_event_listener(handle_new_status_event_wrapper, EventType.STATUS_CLIENT)
        self.controller.add_event_listener(handle_new_desc_event_wrapper, EventType.HS_DESC)
        self.controller.add_event_listener(handle_new_desc_content_event_wrapper, EventType.HS_DESC_CONTENT)

    def shutdown(self):
        self.controller.close()



import sys
import signal
import logging

from onionbalance.common import log

logger = log.get_logger()


class SignalHandler(object):
    """
    Handle signals sent to the OnionBalance daemon process
    """

    def __init__(self, controller, status_socket=None):
        """
        Setup signal handler
        """
        self._tor_controller = controller
        self._status_socket = status_socket

        # Register signal handlers
        signal.signal(signal.SIGTERM, self._handle_sigint_sigterm)
        signal.signal(signal.SIGINT, self._handle_sigint_sigterm)

    def _handle_sigint_sigterm(self, signum, frame):
        """
        Handle SIGINT (Ctrl-C) and SIGTERM

        Disconnect from control port and cleanup the status socket
        """
        logger.info("Signal %d received, exiting", signum)
        self._tor_controller.close()
        if self._status_socket:
            self._status_socket.close()
        logging.shutdown()
        sys.exit(0)



import base64
import bisect
import hashlib

import stem.util

from onionbalance.common import log

from onionbalance.hs_v3 import tor_node
from onionbalance.hs_v3 import params

logger = log.get_logger()


def _time_between_tp_and_srv(valid_after):
    """
     Return True if we are currently in the time segment between a new time
     period and a new SRV (in the real network that happens between 12:00 and
     00:00 UTC). Here is a diagram showing exactly when this returns true:

        +------------------------------------------------------------------+
        |                                                                  |
        | 00:00      12:00       00:00       12:00       00:00       12:00 |
        | SRV#1      TP#1        SRV#2       TP#2        SRV#3       TP#3  |
        |                                                                  |
        |  $==========|-----------$===========|-----------$===========|    |
        |             ^^^^^^^^^^^^            ^^^^^^^^^^^^                 |
        |                                                                  |
        +------------------------------------------------------------------+
    """
    from onionbalance.hs_v3.onionbalance import my_onionbalance

    srv_start_time = my_onionbalance.consensus.get_start_time_of_current_srv_run()
    tp_start_time = my_onionbalance.consensus.get_start_time_of_next_time_period(srv_start_time)
    valid_after = stem.util.datetime_to_unix(valid_after)

    if valid_after >= srv_start_time and valid_after < tp_start_time:
        logger.debug("We are between SRV and TP")
        return False

    logger.debug("We are between TP and SRV (valid_after: %s, srv_start_time: %s -> tp_start_time: %s)",
                 valid_after, srv_start_time, tp_start_time)
    return True


def get_srv_and_time_period(is_first_descriptor):
    """
    Return SRV and time period based on current consensus time
    """
    from onionbalance.hs_v3.onionbalance import my_onionbalance

    valid_after = my_onionbalance.consensus.consensus.valid_after

    current_tp = my_onionbalance.consensus.get_time_period_num()
    previous_tp = current_tp - 1
    next_tp = current_tp + 1
    assert(previous_tp > 0)

    # Get the right TP/SRV
    if is_first_descriptor:
        if _time_between_tp_and_srv(valid_after):
            srv = my_onionbalance.consensus.get_previous_srv(previous_tp)
            tp = previous_tp
            _case = 1  # just for debugging
        else:
            srv = my_onionbalance.consensus.get_previous_srv(current_tp)
            tp = current_tp
            _case = 2  # just for debugging
    else:
        if _time_between_tp_and_srv(valid_after):
            srv = my_onionbalance.consensus.get_current_srv(current_tp)
            tp = current_tp
            _case = 3  # just for debugging
        else:
            srv = my_onionbalance.consensus.get_current_srv(next_tp)
            tp = next_tp
            _case = 4  # just for debugging

    srv_b64 = base64.b64encode(srv) if srv else None
    logger.debug("For valid_after %s we got SRV %s and TP %s (case: #%d)",
                 valid_after, srv_b64, tp, _case)

    return srv, tp


def _get_hash_ring_for_descriptor(is_first_descriptor):
    """
    Return a dictionary { <node hsdir index> : Node , .... }
    """
    from onionbalance.hs_v3.onionbalance import my_onionbalance

    node_hash_ring = {}

    srv, time_period_num = get_srv_and_time_period(is_first_descriptor)
    logger.info("Using srv %s and TP#%s (%s descriptor)",
                srv.hex(), time_period_num,
                "first" if is_first_descriptor else "second")

    for node in my_onionbalance.consensus.nodes:
        try:
            hsdir_index = node.get_hsdir_index(srv, time_period_num)
        except (tor_node.NoEd25519Identity, tor_node.NoHSDir) as e:
            logger.debug("Could not find ed25519 for node %s (%s)", node.routerstatus.fingerprint, e)
            continue

        logger.debug("%s: Node: %s,  index: %s", is_first_descriptor, node.get_hex_fingerprint(), hsdir_index.hex())
        node_hash_ring[hsdir_index] = node

    return node_hash_ring


def _get_hidden_service_index(blinded_pubkey, replica_num, is_first_descriptor):
    """
        hs_index(replicanum) = H("store-at-idx" |
                                 blinded_public_key |
                                 INT_8(replicanum) |
                                 INT_8(period_length) |
                                 INT_8(period_num) )
    """
    from onionbalance.hs_v3.onionbalance import my_onionbalance
    period_length = my_onionbalance.consensus.get_time_period_length()

    replica_num_int_8 = replica_num.to_bytes(8, 'big')
    period_length_int_8 = (period_length).to_bytes(8, 'big')

    _, time_period_num = get_srv_and_time_period(is_first_descriptor)
    logger.info("Getting HS index with TP#%s for %s descriptor (%d replica) ",
                time_period_num,
                "first" if is_first_descriptor else "second", replica_num)
    period_num_int_8 = time_period_num.to_bytes(8, 'big')

    hash_body = b"%s%s%s%s%s" % (b"store-at-idx",
                                 blinded_pubkey,
                                 replica_num_int_8,
                                 period_length_int_8,
                                 period_num_int_8)

    hs_index = hashlib.sha3_256(hash_body).digest()

    return hs_index


def get_responsible_hsdirs(blinded_pubkey, is_first_descriptor):
    """
    Return a list with the responsible HSDirs for a service with 'blinded_pubkey'.

    The returned list is a list of fingerprints.
    """
    from onionbalance.hs_v3.onionbalance import my_onionbalance

    # Always use a live consensus when calculating responsible HSDirs
    assert(my_onionbalance.consensus.is_live())

    responsible_hsdirs = []

    # TODO: Improve representation of hash ring here... No need to go
    # between list and dictionary...

    # dictionary { <node hsdir index> : Node , .... }
    node_hash_ring = _get_hash_ring_for_descriptor(is_first_descriptor)
    if not node_hash_ring:
        raise EmptyHashRing

    sorted_hash_ring_list = sorted(list(node_hash_ring.keys()))

    logger.info("Initialized hash ring of size %d (blinded key: %s)",
                len(node_hash_ring), base64.b64encode(blinded_pubkey))

    for replica_num in range(1, params.HSDIR_N_REPLICAS + 1):
        # The HSDirs that we are gonna store this replica in
        replica_store_hsdirs = []

        hidden_service_index = _get_hidden_service_index(blinded_pubkey, replica_num, is_first_descriptor)

        # Find position of descriptor ID in the HSDir list
        index = bisect.bisect_left(sorted_hash_ring_list, hidden_service_index)

        logger.info("\t Tried with HS index %s got position %d", hidden_service_index.hex(), index)

        while len(replica_store_hsdirs) < params.HSDIR_SPREAD_STORE:
            try:
                hsdir_key = sorted_hash_ring_list[index]
                index += 1
            except IndexError:
                # Wrap around when we reach the end of the HSDir list
                index = 0
                hsdir_key = sorted_hash_ring_list[index]

            hsdir_node = node_hash_ring[hsdir_key]

            # Check if we have already added this node to this
            # replica. This should never happen on the real network but
            # might happen in small testnets like chutney!
            if hsdir_node.get_hex_fingerprint() in replica_store_hsdirs:
                logger.debug("Ignoring already added HSDir to this replica!")
                break

            # Check if we have already added this node to the responsible
            # HSDirs. This can happen in the second replica and we should
            # skip the node
            if hsdir_node.get_hex_fingerprint() in responsible_hsdirs:
                logger.debug("Ignoring already added HSDir!")
                continue

            logger.debug("%d: %s: %s", index, hsdir_node.get_hex_fingerprint(), hsdir_key.hex())

            replica_store_hsdirs.append(hsdir_node.get_hex_fingerprint())

        responsible_hsdirs.extend(replica_store_hsdirs)

    # Do a sanity check
    if my_onionbalance.is_testnet:
        # If we are on chutney it's normal to not have enough nodes to populate the hashring
        assert(len(responsible_hsdirs) <= params.HSDIR_N_REPLICAS * params.HSDIR_SPREAD_STORE)
    else:
        if (len(responsible_hsdirs) != params.HSDIR_N_REPLICAS * params.HSDIR_SPREAD_STORE):
            logger.critical("Got the wrong number of responsible HSDirs: %d. Aborting", len(responsible_hsdirs))
            raise EmptyHashRing

    return responsible_hsdirs


class EmptyHashRing(Exception):
    pass



# -*- coding: utf-8 -*-
import hashlib
import base64
import textwrap
import datetime

import Crypto.Util.number
import stem.descriptor.hidden_service_descriptor

from onionbalance.hs_v2 import util

from onionbalance.common import log
from onionbalance.hs_v2 import config

logger = log.get_logger()


def generate_service_descriptor(permanent_key, introduction_point_list=None,
                                replica=0, timestamp=None, deviation=0):
    """
    High-level interface for generating a signed HS descriptor
    """

    if not timestamp:
        timestamp = datetime.datetime.utcnow()
    unix_timestamp = int(timestamp.strftime("%s"))

    permanent_key_block = make_public_key_block(permanent_key)
    permanent_id = util.calc_permanent_id(permanent_key)

    # Calculate the current secret-id-part for this hidden service
    # Deviation allows the generation of a descriptor for a different time
    # period.
    time_period = (util.get_time_period(unix_timestamp, permanent_id) + int(deviation))

    secret_id_part = util.calc_secret_id_part(time_period, None, replica)
    descriptor_id = util.calc_descriptor_id(permanent_id, secret_id_part)

    if not introduction_point_list:
        onion_address = util.calc_onion_address(permanent_key)
        raise ValueError("No introduction points for service %s.onion." %
                         onion_address)

    # Generate the introduction point section of the descriptor
    intro_section = make_introduction_points_part(
        introduction_point_list
    )

    unsigned_descriptor = generate_hs_descriptor_raw(
        desc_id_base32=util.base32_encode_str(descriptor_id),
        permanent_key_block=permanent_key_block,
        secret_id_part_base32=util.base32_encode_str(secret_id_part),
        publication_time=util.rounded_timestamp(timestamp),
        introduction_points_part=intro_section
    )

    signed_descriptor = sign_descriptor(unsigned_descriptor, permanent_key)
    return signed_descriptor


def generate_hs_descriptor_raw(desc_id_base32, permanent_key_block,
                               secret_id_part_base32, publication_time,
                               introduction_points_part):
    """
    Generate hidden service descriptor string
    """
    doc = [
        "rendezvous-service-descriptor {}".format(desc_id_base32),
        "version 2",
        "permanent-key",
        permanent_key_block,
        "secret-id-part {}".format(secret_id_part_base32),
        "publication-time {}".format(publication_time),
        "protocol-versions 2,3",
        "introduction-points",
        introduction_points_part,
        "signature\n",
    ]

    unsigned_descriptor = '\n'.join(doc)
    return unsigned_descriptor


def make_introduction_points_part(introduction_point_list=None):
    """
    Make introduction point block from list of IntroductionPoint objects
    """

    # If no intro points were specified, we should create an empty list
    if not introduction_point_list:
        introduction_point_list = []

    intro = []
    for intro_point in introduction_point_list:
        intro.append("introduction-point {}".format(intro_point.identifier))
        intro.append("ip-address {}".format(intro_point.address))
        intro.append("onion-port {}".format(intro_point.port))
        intro.append("onion-key")
        intro.append(intro_point.onion_key)
        intro.append("service-key")
        intro.append(intro_point.service_key)

    intro_section = '\n'.join(intro).encode('utf-8')
    intro_section_base64 = base64.b64encode(intro_section).decode('utf-8')
    intro_section_base64 = textwrap.fill(intro_section_base64, 64)

    # Add the header and footer:
    intro_points_with_headers = '\n'.join([
        '-----BEGIN MESSAGE-----',
        intro_section_base64,
        '-----END MESSAGE-----'])
    return intro_points_with_headers


def make_public_key_block(key):
    """
    Get ASN.1 representation of public key, base64 and add headers
    """
    asn1_pub = util.get_asn1_sequence(key)
    pub_base64 = base64.b64encode(asn1_pub).decode('utf-8')
    pub_base64 = textwrap.fill(pub_base64, 64)

    # Add the header and footer:
    pub_with_headers = '\n'.join([
        '-----BEGIN RSA PUBLIC KEY-----',
        pub_base64,
        '-----END RSA PUBLIC KEY-----'])
    return pub_with_headers


def sign_digest(digest, private_key):
    """
    Sign, base64 encode, wrap and add Tor signature headers

    The message digest is PKCS1 padded without the optional
    algorithmIdentifier section.
    """

    digest = util.add_pkcs1_padding(digest)
    (signature_long, ) = private_key.sign(digest, None)
    signature_bytes = Crypto.Util.number.long_to_bytes(signature_long, 128)
    signature_base64 = base64.b64encode(signature_bytes).decode('utf-8')
    signature_base64 = textwrap.fill(signature_base64, 64)

    # Add the header and footer:
    signature_with_headers = '\n'.join([
        '-----BEGIN SIGNATURE-----',
        signature_base64,
        '-----END SIGNATURE-----'])
    return signature_with_headers


def sign_descriptor(descriptor, service_key):
    """
    Sign or resign a provided hidden service descriptor
    """
    token_descriptor_signature = '\nsignature\n'

    # Remove signature block if it exists
    if token_descriptor_signature in descriptor:
        descriptor = descriptor[:descriptor.find(token_descriptor_signature) + len(token_descriptor_signature)]
    else:
        descriptor = descriptor.strip() + token_descriptor_signature

    descriptor_digest = hashlib.sha1(descriptor.encode('utf-8')).digest()
    signature_with_headers = sign_digest(descriptor_digest, service_key)
    return descriptor + signature_with_headers


def descriptor_received(descriptor_content):
    """
    Process onion service descriptors retrieved from the HSDir system or
    received directly over the metadata channel.
    """

    try:
        parsed_descriptor = stem.descriptor.hidden_service_descriptor.\
            HiddenServiceDescriptor(descriptor_content, validate=True)
    except ValueError:
        logger.exception("Received an invalid service descriptor.")
        return None

    # Ensure the received descriptor matches the requested descriptor
    permanent_key = Crypto.PublicKey.RSA.importKey(
        parsed_descriptor.permanent_key)
    descriptor_onion_address = util.calc_onion_address(permanent_key)

    known_descriptor, instance_changed = False, False
    for instance in [instance for service in config.services for
                     instance in service.instances]:
        if instance.onion_address == descriptor_onion_address:
            instance_changed |= instance.update_descriptor(parsed_descriptor)
            known_descriptor = True

    if instance_changed:
        logger.info("The introduction point set has changed for instance "
                    "%s.onion.", descriptor_onion_address)

    if not known_descriptor:
        # No matching service instance was found for the descriptor
        logger.debug("Received a descriptor for an unknown service:\n%s",
                     descriptor_content.decode('utf-8'))
        logger.warning("Received a descriptor with address %s.onion that "
                       "did not match any configured service instances.",
                       descriptor_onion_address)

    return None



# -*- coding: utf-8 -*-

"""setup.py: setuptools control."""

import io
import versioneer
import os

from setuptools import setup

# Read version and other info from package's __init.py file
module_info = {}
init_path = os.path.join(os.path.dirname(__file__), 'onionbalance',
                         '__init__.py')
with open(init_path) as init_file:
    exec(init_file.read(), module_info)


def read(*names, **kwargs):
    return io.open(
        os.path.join(os.path.dirname(__file__), *names),
        encoding=kwargs.get("encoding", "utf8")
    ).read()

setup(
    name="OnionBalance",
    packages=["onionbalance",
              "onionbalance.hs_v2", "onionbalance.hs_v3", "onionbalance.hs_v3.ext", "onionbalance.common",
              "onionbalance.config_generator"],
    entry_points={
        "console_scripts": [
            'onionbalance = onionbalance.manager:main',
            'onionbalance-config = onionbalance.config_generator.config_generator:main',
        ]},
    description="OnionBalance provides load-balancing and redundancy for Tor "
                "hidden services by distributing requests to multiple backend "
                "Tor instances.",
    long_description=read('README.rst'),
    author=module_info.get('__author__'),
    author_email=module_info.get('__contact__'),
    url=module_info.get('__url__'),
    license=module_info.get('__license__'),
    keywords='tor',
    python_requires='>=3.6',
    install_requires=[
        'setuptools',
        'stem>=1.8',
        'PyYAML>=4.2b1',
        'pycrypto>=2.6.1',
        'future>=0.14.0',
        'setproctitle',
        'cryptography>=2.5',
        ],
    tests_require=['tox', 'pytest-mock', 'pytest', 'mock', 'pexpect'],
    package_data={'onionbalance.config_generator': ['data/*']},
    include_package_data=True,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: GNU General Public License (GPL)',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
    ],
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
)



import logging

from onionbalance.common import scheduler
from onionbalance.common import log
from onionbalance.common import signalhandler

from onionbalance.hs_v3 import params
from onionbalance.hs_v3 import onionbalance

logger = log.get_logger()


def main(args):
    """
    This is the entry point of v3 functionality.

    Initialize onionbalance, schedule future jobs and let the scheduler do its thing.
    """

    # Override the log level if specified on the command line.
    if args.verbosity:
        params.DEFAULT_LOG_LEVEL = args.verbosity.upper()
    logger.setLevel(logging.__dict__[params.DEFAULT_LOG_LEVEL.upper()])

    # Get the global onionbalance singleton
    my_onionbalance = onionbalance.my_onionbalance
    my_onionbalance.init_subsystems(args)

    signalhandler.SignalHandler(my_onionbalance.controller.controller)

    # Schedule descriptor fetch and upload events
    if my_onionbalance.is_testnet:
        scheduler.add_job(params.FETCH_DESCRIPTOR_FREQUENCY_TESTNET, my_onionbalance.fetch_instance_descriptors)
        scheduler.add_job(params.PUBLISH_DESCRIPTOR_CHECK_FREQUENCY_TESTNET, my_onionbalance.publish_all_descriptors)
    else:
        scheduler.add_job(params.FETCH_DESCRIPTOR_FREQUENCY, my_onionbalance.fetch_instance_descriptors)
        scheduler.add_job(params.PUBLISH_DESCRIPTOR_CHECK_FREQUENCY, my_onionbalance.publish_all_descriptors)

    # Run initial fetch of HS instance descriptors
    scheduler.run_all(delay_seconds=params.INITIAL_CALLBACK_DELAY)

    # Begin main loop to poll for HS descriptors
    scheduler.run_forever()

    return 0



# This is the ed25519 implementation from
#     http://ed25519.cr.yp.to/python/ed25519.py .
# It is in the public domain.
#
# It isn't constant-time.  Don't use it except for testing.  Also, see
# warnings about how very slow it is.  Only use this for generating
# test vectors, I'd suggest.
#
# Don't edit this file.  Mess with ed25519_ref.py

import hashlib

b = 256
q = 2**255 - 19
l = 2**252 + 27742317777372353535851937790883648493

def H(m):
  return hashlib.sha512(m).digest()

def expmod(b,e,m):
  if e == 0: return 1
  t = expmod(b,e//2,m)**2 % m
  if e & 1: t = (t*b) % m
  return t

def inv(x):
  return expmod(x,q-2,q)

d = -121665 * inv(121666)
I = expmod(2,(q-1)//4,q)

def xrecover(y):
  xx = (y*y-1) * inv(d*y*y+1)
  x = expmod(xx,(q+3)//8,q)
  if (x*x - xx) % q != 0: x = (x*I) % q
  if x % 2 != 0: x = q-x
  return x

By = 4 * inv(5)
Bx = xrecover(By)
B = [Bx % q,By % q]

def edwards(P,Q):
  x1 = P[0]
  y1 = P[1]
  x2 = Q[0]
  y2 = Q[1]
  x3 = (x1*y2+x2*y1) * inv(1+d*x1*x2*y1*y2)
  y3 = (y1*y2+x1*x2) * inv(1-d*x1*x2*y1*y2)
  return [x3 % q,y3 % q]

def scalarmult(P,e):
  if e == 0: return [0,1]
  Q = scalarmult(P,e//2)
  Q = edwards(Q,Q)
  if e & 1: Q = edwards(Q,P)
  return Q

def encodeint(y):
  bits = [(y >> i) & 1 for i in range(b)]
  return b''.join([bytes([sum([bits[i * 8 + j] << j for j in range(8)])]) for i in range(b//8)])

def encodepoint(P):
  x = P[0]
  y = P[1]
  bits = [(y >> i) & 1 for i in range(b - 1)] + [x & 1]
  return b''.join([bytes([sum([bits[i * 8 + j] << j for j in range(8)])]) for i in range(b//8)])

def bit(h,i):
  return (h[i//8] >> (i%8)) & 1

def publickey(sk):
  h = H(sk)
  a = 2**(b-2) + sum(2**i * bit(h,i) for i in range(3,b-2))
  A = scalarmult(B,a)
  return encodepoint(A)

def Hint(m):
  h = H(m)
  return sum(2**i * bit(h,i) for i in range(2*b))

def signature(m,sk,pk):
  h = H(sk)
  a = 2**(b-2) + sum(2**i * bit(h,i) for i in range(3,b-2))
  r = Hint(b''.join([bytes([h[i]]) for i in range(b//8,b//4)]) + m)
  R = scalarmult(B,r)
  S = (r + Hint(encodepoint(R) + pk + m) * a) % l
  return encodepoint(R) + encodeint(S)

def isoncurve(P):
  x = P[0]
  y = P[1]
  return (-x*x + y*y - 1 - d*x*x*y*y) % q == 0

def decodeint(s):
  return sum(2**i * bit(s,i) for i in range(0,b))

def decodepoint(s):
  y = sum(2**i * bit(s,i) for i in range(0,b-1))
  x = xrecover(y)
  if x & 1 != bit(s,b-1): x = q-x
  P = [x,y]
  if not isoncurve(P): raise Exception("decoding point that is not on curve")
  return P

def checkvalid(s,m,pk):
  if len(s) != b//4: raise Exception("signature length is wrong")
  if len(pk) != b//8: raise Exception("public-key length is wrong")
  R = decodepoint(s[0:b//8])
  A = decodepoint(pk)
  S = decodeint(s[b//8:b//4])
  h = Hint(encodepoint(R) + pk + m)
  if scalarmult(B,S) != edwards(R,scalarmult(A,h)):
    raise Exception("signature does not pass verification")



# -*- coding: utf-8 -*-
import datetime

from onionbalance.common import log
from onionbalance.hs_v2 import config
import onionbalance.common.instance

logger = log.get_logger()


def fetch_instance_descriptors(controller):
    all_instances = [instance for service in config.services for instance in service.instances]

    onionbalance.common.instance.helper_fetch_all_instance_descriptors(controller, all_instances,
                                                                       control_password=config.TOR_CONTROL_PASSWORD)


class InstanceV2(onionbalance.common.instance.Instance):
    """
    This is a V2 onionbalance instance
    """

    def __init__(self, controller, onion_address, authentication_cookie):
        """
        Constructor for V2 instance
        """
        # Initialize the common instance class
        super().__init__(controller, onion_address)

        self.authentication_cookie = authentication_cookie

        # Store the latest set of introduction points for this instance
        self.introduction_points = []

        # Timestamp when last received a descriptor for this instance
        self.received = None

        # Timestamp of the currently loaded descriptor
        self.timestamp = None

    def update_descriptor(self, parsed_descriptor):
        """
        Update introduction points when a new HS descriptor is received

        Parse the descriptor content and update the set of introduction
        points for this HS instance. Returns True if the introduction
        point set has changed, False otherwise.`
        """

        self.received = datetime.datetime.utcnow()

        logger.debug("Received a descriptor for instance %s.onion.",
                     self.onion_address)

        # Reject descriptor if its timestamp is older than the current
        # descriptor. Prevents HSDirs from replaying old, expired
        # descriptors.
        if self.timestamp and parsed_descriptor.published < self.timestamp:
            logger.error("Received descriptor for instance %s.onion with "
                         "publication timestamp (%s) older than the latest "
                         "descriptor (%s). Ignoring the descriptor.",
                         self.onion_address,
                         parsed_descriptor.published,
                         self.timestamp)
            return False
        else:
            self.timestamp = parsed_descriptor.published

        # Parse the introduction point list, decrypting if necessary
        introduction_points = parsed_descriptor.introduction_points(
            authentication_cookie=self.authentication_cookie
        )

        # If the new introduction points are different, flag this instance
        # as modified. Compare the set of introduction point identifiers
        # (fingerprint of the per IP circuit service key).
        if (set(ip.identifier for ip in introduction_points) != set(ip.identifier for ip in self.introduction_points)):
            self.intro_set_changed_since_published = True
            self.introduction_points = introduction_points
            return True

        else:
            logger.debug("Introduction points for instance %s.onion matched "
                         "the cached set.", self.onion_address)
            return False
